/* vim:sw=4 ts=8 tw=0 noai cin cino=>1s+0.5s\:0t0
 *
 * pam_krb5_auth.c
 *
 * PAM authentication management functions for pam_krb5
 *
 */

static const char rcsid[] = "$Id: pam_krb5_auth.c,v 1.1.1.1.2.14 2001/09/24 20:19:45 nectar Exp $";

#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <errno.h>
#include <limits.h>	/* PATH_MAX */
#include <pwd.h>	/* getpwnam */
#include <stdio.h>	/* snprintf */
#include <stdlib.h>	/* malloc  */
#include <strings.h>	/* strchr */
#include <string.h>	/* strstr */
#include <syslog.h>	/* syslog */
#include <unistd.h>	/* chown */

#include <krb5.h>
#include <com_err.h>
#include "pam_krb5.h"


#define DEFAULT_RENEW_LIFE (60 * 60 * 24 * 5)


static void	set_gic_options(struct pam_krb5_state *,
		    enum pam_krb5_get_cred_type, krb5_get_init_creds_opt *);
static int	resolve_principal(struct pam_krb5_state *);
static int	verify_local_name(struct pam_krb5_state *);
static void	cleanup(struct pam_krb5_state *);
static int	get_user_password(struct pam_krb5_state *, char **);

/* Authenticate a user via krb5 */
int
pam_sm_authenticate(pam_handle_t *pamh, int flags, int argc,
    const char **argv)
{
    struct		 pam_krb5_state *state;
    krb5_creds		*creds;
    int			 e;

    e = pam_krb5_initialize_method("pam_sm_authenticate", pamh, &state,
	argc, argv);
    if (e != PAM_SUCCESS)
	return e;

    CLEAR_STATE(state, STATE_WILL_IGNORE);
    SET_STATE(state, STATE_SM_AUTHENTICATE);
    creds = NULL;
    
    e = resolve_principal(state);
    if (e == PAM_SUCCESS)
	e = pam_krb5_get_authtok(state, CRED_TGT, &creds);
    if (e == PAM_SUCCESS)
	e = verify_local_name(state);
    if (e == PAM_SUCCESS && creds != NULL)
    	e = pam_krb5_store_tgt(state, creds);
    if (e == PAM_SUCCESS && creds != NULL)
	e = pam_krb5_verify_tgt(state);
    if (e == PAM_SUCCESS && creds != NULL &&
	TEST_OPTION(state, OPT_SETCRED_IN_AUTH))
	e = pam_krb5_setcred(state, PAM_ESTABLISH_CRED);
    if (creds != NULL)
	krb5_free_creds(state->ctx, creds);
    if (e == PAM_SUCCESS)
	SET_STATE(state, STATE_AUTH_COMPLETED);
    else
	cleanup(state);
    pam_krb5_note_result("pam_sm_authenticate", state, e);
    CLEAR_STATE(state, STATE_SM_AUTHENTICATE);
    return e;
}

static void
cleanup(struct pam_krb5_state *state)
{

    FREE0(state->princ_name);
    if (state->ctx) {
	if (state->princ) {
	    krb5_free_principal(state->ctx, state->princ);
	    state->princ = NULL;
	}
	if (state->ccache) {
	    (void)krb5_cc_destroy(state->ctx, state->ccache);
	    state->ccache = NULL;
	}
	if (state->chpw_creds) {
	    krb5_free_creds(state->ctx, state->chpw_creds);
	    state->chpw_creds = NULL;
	}
    }
}

static void
set_gic_options(struct pam_krb5_state *state, enum pam_krb5_get_cred_type type,
    krb5_get_init_creds_opt *gic_opt)
{
    krb5_error_code	 e;
    krb5_deltat		 rlife;
    const char		*renew;

    krb5_get_init_creds_opt_init(gic_opt);
    switch (type) {
    case CRED_TGT:
    	renew = state->opt.renewable;
	krb5_get_init_creds_opt_set_forwardable(gic_opt,
	    TEST_OPTION(state, OPT_FORWARDABLE) ? 1 : 0);
	if (TEST_OPTION(state, OPT_RENEWABLE))
	    krb5_get_init_creds_opt_set_renew_life(gic_opt,
		DEFAULT_RENEW_LIFE);
	if (renew != NULL) {
	    if ((e = krb5_string_to_deltat(renew, &rlife)) == 0)
		krb5_get_init_creds_opt_set_renew_life(gic_opt, rlife);
	    else
	    DBG(("pam_sm_authenticate: couldn't parse renewable date %s: %s",
		    renew, error_message(e)));
	}
	break;
    case CRED_CHANGEPW:
	krb5_get_init_creds_opt_set_tkt_life(gic_opt, 60);
	krb5_get_init_creds_opt_set_forwardable(gic_opt, 0);
	krb5_get_init_creds_opt_set_proxiable(gic_opt, 0);
	break;
    default:
	LOG((LOG_CRIT, "set_gic_options: unexpected cred type %d"));
	break;
    }
}

static int
resolve_principal(struct pam_krb5_state *state)
{
    int			e;
    krb5_error_code	krbret;

    /* We'll reset pamret if anything goes wrong */
    e = PAM_SUCCESS;

    /* Special handling for username `root' */
    if (strcmp(state->user, "root") == 0) {
	if (!TEST_OPTION(state, OPT_ROOTOK))
	    /* If `rootok' was not specified, then we will ultimately
	     * return PAM_IGNORE.  Not yet, though, because we want
	     * to prompt for a password.
	     */ 
	    SET_STATE(state, STATE_WILL_IGNORE);	    
	if (TEST_OPTION(state, OPT_NOROOT))
	    /* If `noroot' was specified, then we want to return
	     * PAM_AUTH_ERR. XXX Shouldn't we prompt for a password
	     * in this case as well?
	     */
	    e = PAM_AUTH_ERR;
	else if ((krbret = krb5_sname_to_principal(state->ctx, NULL, "root",
	    KRB5_NT_SRV_HST, &state->princ)) != 0) {
	    /* XXX How is this useful?  Shouldn't we be using
	     * root@REALM here?
	     */
	    /* Needed for SEAM/RPCSEC_GSS/Secure NFS */
	    DBG(("pam_sm_authenitcate: krb5_sname_to_principal: %s",
	    	error_message(krbret)));
	    e = PAM_SERVICE_ERR;
	}

    /* This is the usual case */
    } else if ((krbret = krb5_parse_name(state->ctx, state->user,
	&state->princ)) != 0) {
	DBG(("pam_sm_authenticate: krb5_parse_name `%s': %s", state->user,
	    error_message(krbret)));
	e = PAM_SERVICE_ERR;
    }

    /* Now convert the principal name into something human readable */
    if (e == PAM_SUCCESS && (krbret = krb5_unparse_name(state->ctx,
	state->princ, &state->princ_name)) != 0) {
	DBG(("pam_sm_authenticate: krb5_unparse_name(%s)",
	    error_message(krbret)));
	e = PAM_SERVICE_ERR;
    }
    DBG(("pam_sm_authenticate: resolve_principal: %s",
	pam_strerror(state->pamh, e)));
    return e;
}

int
pam_krb5_get_authtok(struct pam_krb5_state *state,
    enum pam_krb5_get_cred_type type, krb5_creds **creds)
{
    char	*pass;
    int		 e, badpass;
    
    *creds  = NULL;
    pass    = NULL;
    badpass = 0;

    e = PAM_SUCCESS;
    if (TEST_OPTION(state, OPT_USE_FIRST_PASS | OPT_TRY_FIRST_PASS)) {
	e = pam_get_item(state->pamh, PAM_AUTHTOK, (PAM_CONST void **)&pass);
	if (pass != NULL)
	    e = pam_krb5_get_cred(state, pass, type, creds, &badpass);
	else {
	    if (e == PAM_SUCCESS)
		DBG(("pam_krb5_get_authtok: no pre-existing password"));
	    else
	    DBG(("pam_krb5_get_authtok: pam_get_item PAM_AUTHTOK failed: %s",
	    	    pam_strerror(state->pamh, e)));
	}
    }
    if (*creds == NULL &&
	(pass == NULL || 
	    (TEST_OPTION(state, OPT_TRY_FIRST_PASS) && badpass))) {
	e = get_user_password(state, &pass);
	if (e == PAM_SUCCESS)
	    e = pam_krb5_get_cred(state, pass, type, creds, &badpass);
    }
    if (TEST_STATE(state, STATE_WILL_IGNORE)) {
    DBG(("pam_krb5_get_authtok: because STATE_WILL_IGNORE, `%s' becomes `%s'",
	pam_strerror(state->pamh, e), pam_strerror(state->pamh, PAM_IGNORE)));
	e = PAM_IGNORE;
    }
    return e;
}

int
pam_krb5_get_cred(struct pam_krb5_state *state, char *pass,
    enum pam_krb5_get_cred_type type, krb5_creds **creds, int *badpass)
{
    krb5_error_code		 e;
    int			 	 epam;
    const char			*service;
    krb5_get_init_creds_opt	 gic_opt, chpw_opt;

    set_gic_options(state, type, &gic_opt);
    if (type == CRED_CHANGEPW)
	service = "kadmin/changepw";
    else
	service = NULL;
    if (*creds == NULL)
	*creds = malloc(sizeof(**creds));
    if (*creds == NULL) {
	LOG((LOG_CRIT, "get_cred: out of memory"));
    	return PAM_SYSTEM_ERR;
    }
    e = krb5_get_init_creds_password(state->ctx, *creds, state->princ, pass,
	pam_krb5_prompter, state, 0, service, &gic_opt);
    DBG(("pam_sm_authenticate: krb5_get_init_creds_password: %s",
	e == 0 ? "Success" : error_message(e)));
    switch (e) {
    case 0:
	epam = PAM_SUCCESS;
	break;
    case KRB5KRB_AP_ERR_BAD_INTEGRITY:
	FREE0(*creds);
	if (badpass != NULL)
	    *badpass = 1;
	epam = PAM_AUTH_ERR;
	break;
    case KRB5KDC_ERR_C_PRINCIPAL_UNKNOWN:
	FREE0(*creds);
	if (TEST_OPTION(state, OPT_IGNORE_UNKNOWN_UPN)) {
	    LOG((LOG_CRIT, "pam_sm_authenticate: ignore_unknown_upn: %s",
		state->princ_name));
	    epam = PAM_IGNORE;
	} else
	    epam = PAM_USER_UNKNOWN;
	break;
    case KRB5KDC_ERR_KEY_EXP:
	/* KRB5KDC_ERR_KEY_EXP will only be returned by 
	 * krb5_get_init_creds_password if pam_krb5_prompter returns
	 * it.  pam_krb5_prompter will only return KRB5KDC_ERR_KEY_EXP
	 * if OPT_PW_EXP_IN_AUTH is not set.  This implies that we only
	 * take this code path if OPT_PW_EXP_IN_AUTH is not set.
	 */
	/* It isn't expected that we get this result if we were getting
	 * credentials for changepw.
	 */
	if (TEST_OPTION(state, OPT_ERR_ON_EXP) || type == CRED_CHANGEPW)
	    epam = PAM_AUTH_ERR;
	else {
	    SET_STATE(state, STATE_PW_EXPIRED);
	    state->chpw_creds = *creds;
	    *creds = NULL;
	    set_gic_options(state, CRED_CHANGEPW, &chpw_opt);
	    if (gic_opt.flags & KRB5_GET_INIT_CREDS_OPT_PREAUTH_LIST)
		krb5_get_init_creds_opt_set_preauth_list(&chpw_opt,
		    gic_opt.preauth_list, gic_opt.preauth_list_length);
	    e = krb5_get_init_creds_password(state->ctx,
		state->chpw_creds, state->princ, pass, NULL, state, 0,
		"kadmin/changepw", &chpw_opt);
	    if (e == 0) {
		DBG(("pam_krb5_authtok: got chpw_creds"));
		epam = PAM_SUCCESS;
	    } else {
		DBG(("pam_krb5_authtok: krb5_get_init_creds_password: %s",
		    error_message(e)));
		FREE0(state->chpw_creds);
		epam = PAM_AUTH_ERR;
	    }
	}
	break;
    default:
	FREE0(*creds);
	epam = PAM_AUTH_ERR;
	break;
    }
    return epam;

}

static int
get_user_password(struct pam_krb5_state *state, char **pass)
{
    char	 prompt[80];
    char	*p;
    int		 e;

    *pass = NULL;
    if (TEST_OPTION(state, OPT_OLDPROMPTS))
	(void)snprintf(prompt, sizeof(prompt), "Password: ");
    else
	(void)snprintf(prompt, sizeof(prompt), "Password for %s: ",
		       state->princ_name);
    e = pam_krb5_get_user_info(state->pamh, prompt, PAM_PROMPT_ECHO_OFF, pass);
    if (e != PAM_SUCCESS)
	DBG(("pam_sm_authenticate: pam_krb5_get_user_info: %s",
		    pam_strerror(state->pamh, e)));
    if (e == PAM_SUCCESS) {
	e = pam_set_item(state->pamh, PAM_AUTHTOK, *pass);
	for (p = *pass; *p; p++)
	    *p = '\0';
	FREE0(*pass);
	if (e != PAM_SUCCESS)
	    DBG(("pam_sm_authenticate: pam_set_item: PAM_AUTHTOK: %s",
			pam_strerror(state->pamh, e)));
    }
    if (e == PAM_SUCCESS)
	e = pam_get_item(state->pamh, PAM_AUTHTOK, (PAM_CONST void **)pass);
    return e;
}

static int
verify_local_name(struct pam_krb5_state *state)
{
    char	lname[80];
    struct	passwd *pw;
    long	krbret;
    int		pamret;
    
    if (strchr(state->user, '@')) {
	krbret = krb5_aname_to_localname(state->ctx, state->princ,
	    sizeof(lname), lname);
	if (krbret != 0) {
	    DBG(("pam_sm_authenticate: krb5_aname_to_localname: %s",
		error_message(krbret)));
	    return PAM_USER_UNKNOWN;
	}
	DBG(("pam_sm_authenticate: changing PAM_USER to `%s'", lname));
	pamret = pam_set_item(state->pamh, PAM_USER, lname);
	if (pamret != PAM_SUCCESS) {
	    DBG(("pam_sm_authenticate: pam_set_item: %s",
		pam_strerror(state->pamh, pamret)));
	    return PAM_SERVICE_ERR;
	}
	pamret = pam_get_item(state->pamh, PAM_USER,
	    (PAM_CONST void **)&state->user);
	if (pamret != PAM_SUCCESS) {
	    DBG(("pam_sm_authenticate: pam_get_item: %s",
		pam_strerror(state->pamh, pamret)));
	    return PAM_SERVICE_ERR;
	}
    }
    pw = getpwnam(state->user);
    if (pw == NULL) {
	DBG(("pam_sm_authenticate: getpwnam failed: %s", state->user));
	return PAM_SUCCESS;
    }

    /* Deal with accounts with uid = 0 */
    if (pw->pw_uid == 0 && !TEST_OPTION(state, OPT_ROOTOK)) {
	if (TEST_OPTION(state, OPT_NOROOT))
	    return PAM_AUTH_ERR;
	else
	    return PAM_IGNORE;
    }
    return PAM_SUCCESS;
}

int
pam_krb5_store_tgt(struct pam_krb5_state *state, krb5_creds *creds)
{
    krb5_ccache		ccache;
    krb5_error_code	e;
    int			epam;

    /* Generate a unique cache */
    e = compat_new_memory_cache(state->ctx, &ccache);
    if (e != 0) {
	DBG(("pam_sm_authenticate: compat_new_memory_cache (%s)", 
	    error_message(e)));
	epam = PAM_SERVICE_ERR;
    }
    if (e == 0) {
	e = krb5_cc_initialize(state->ctx, ccache, state->princ);
	if (e != 0) {
	    DBG(("pam_sm_authenticate: krb5_cc_initialize(%s)",
		error_message(e)));
	    epam = PAM_SERVICE_ERR;
	}
    }
    if (e == 0) {
	e = krb5_cc_store_cred(state->ctx, ccache, creds);
	if (e != 0) {
	    DBG(("pam_sm_authenticate: krb5_cc_store_cred(%s)",
		error_message(e)));
	    (void)krb5_cc_destroy(state->ctx, ccache);
	    epam = PAM_SERVICE_ERR;
	}
    }
    if (e == 0) {
	if (state->ccache != NULL) {
	    DBG(("pam_sm_authenticate: ccache already present"));
	    (void)krb5_cc_destroy(state->ctx, state->ccache);
	}
	state->ccache = ccache;
	epam = PAM_SUCCESS;
    }
    DBG(("pam_sm_authenticate: pam_krb5_store_tgt: %s",
	    pam_strerror(state->pamh, epam)));
    return epam;
}
