/*
 * pam_krb5_auth.c
 *
 * PAM authentication management functions for pam_krb5
 *
 */

static const char rcsid[] = "$Id: pam_krb5_cred.c,v 1.1.2.16 2001/09/24 20:24:26 vorlon Exp $";

#include <sys/types.h>
#include <sys/stat.h>
#include <errno.h>
#include <limits.h>	/* PATH_MAX */
#include <pwd.h>	/* getpwnam */
#include <stdio.h>	/* snprintf */
#include <stdlib.h>	/* malloc  */
#include <strings.h>	/* strchr */
#include <string.h>	/* strstr */
#include <syslog.h>	/* syslog */
#include <unistd.h>	/* chown */

#include <krb5.h>
#include <com_err.h>
#include "pam_krb5.h"

static int	establish_cred(struct pam_krb5_state *);
static int	delete_cred(struct pam_krb5_state *);
static int	reinitialize_cred(struct pam_krb5_state *);
static int	refresh_cred(struct pam_krb5_state *);
static void	expand_ccache_spec(const char *, struct passwd *, char **);

/* Called after a successful authentication. Set user credentials. */
int
pam_sm_setcred(pam_handle_t *pamh, int flags, int argc,
	       const char **argv)
{
    int		pamret;
    struct	pam_krb5_state *state;

    pamret = pam_krb5_initialize_method("pam_sm_setcred", pamh, &state,
	argc, argv);
    if (pamret == PAM_SUCCESS)
    	pamret = pam_krb5_setcred(state, flags);
    pam_krb5_note_result("pam_sm_setcred", state, pamret);
    return pamret;
}

int
pam_krb5_setcred(struct pam_krb5_state *state, int flags)
{
    int	pamret;

    DBG(("pam_sm_setcred: entry"));

    if (!TEST_STATE(state, STATE_AUTH_COMPLETED | STATE_SM_AUTHENTICATE) ||
	TEST_STATE(state, STATE_WILL_IGNORE))
	return PAM_IGNORE;

    pamret = PAM_SUCCESS;
    if (flags & PAM_ESTABLISH_CRED)
	pamret = establish_cred(state);
    if (flags & PAM_DELETE_CRED)
	pamret = delete_cred(state);
    if (flags & PAM_REINITIALIZE_CRED)
	pamret = reinitialize_cred(state);
    if (flags & PAM_REFRESH_CRED)
	pamret = refresh_cred(state);
    DBG(("pam_krb5_setcred: %s", pam_strerror(state->pamh, pamret)));
    return pamret;
}

static int
establish_cred(struct pam_krb5_state *state)
{
    krb5_error_code	 krbret;
    krb5_ccache		 ccache_perm;
    int			 pamret;
    const char		*cache_type, *cache_name;
    struct		 passwd	*pw = NULL;
    char		 buf[PATH_MAX];

    DBG(("pam_sm_setcred: establishing credentials"));
    if (state->ccache == NULL) {
	/* No ccache found.  This might happen if our password expired
	 * and the application didn't call pam_sm_chauthtok. 
	 */
	DBG(("pam_sm_setcred: no ccache found"));
	return PAM_SUCCESS;
    }
    if (TEST_OPTION(state, OPT_NO_CCACHE)) {
	DBG(("pam_sm_setcred: no_ccache option specified"));
	return PAM_SUCCESS;
    }

    /* Get the uid. This should exist. */
    pw = getpwnam(state->user);
    if (pw == NULL) {
	DBG(("pam_sm_setcred: getpwnam failed: %s", state->user));
	return PAM_USER_UNKNOWN;
    }

    /* Avoid following a symlink as root */
    pamret = pam_krb5_become_user(state, pw);
    if (pamret != PAM_SUCCESS)
	return pamret;
    if (state->opt.ccache_spec == NULL)
	krbret = krb5_cc_default(state->ctx, &ccache_perm);
    else if (strcmp(state->opt.ccache_spec, "SAFE") == 0) {
	/* ccache=SAFE */
	krbret = compat_new_file_cache(state->ctx, &ccache_perm);
    } else {
	/* ccache=<cache spec> */
	char *p;

	expand_ccache_spec(state->opt.ccache_spec, pw, &p);
	krbret = krb5_cc_resolve(state->ctx, p, &ccache_perm);
	free(p);
    }
    if (krbret != 0) {
	DBG(("pam_sm_setcred: krb5_cc_resolve: %s", error_message(krbret)));
	pamret = PAM_SERVICE_ERR;
    }
    if (pamret == PAM_SUCCESS) {
	krbret = krb5_cc_initialize(state->ctx, ccache_perm, state->princ);
	if (krbret != 0) {
	    DBG(("pam_sm_setcred: krb5_cc_initialize: %s",
		error_message(krbret)));
	    pamret = PAM_SERVICE_ERR;
	}
    }
    if (pamret == PAM_SUCCESS) {
	cache_type = krb5_cc_get_type(state->ctx, ccache_perm);
    	cache_name = krb5_cc_get_name(state->ctx, ccache_perm);
    	krbret = krb5_cc_copy_cache(state->ctx, state->ccache, ccache_perm);
	(void)krb5_cc_destroy(state->ctx, state->ccache);
	state->ccache = NULL;
    	if (krbret != 0) {
	    DBG(("pam_sm_setcred: krb5_cc_copy_creds: %s",
		error_message(krbret)));
	    pamret = PAM_SERVICE_ERR;
	} 
    }
    (void)pam_krb5_unbecome_user(state);
    if (pamret == PAM_SUCCESS && strcmp(cache_type, "FILE") == 0) {
	/* This is probably just paranoia, but better safe than sorry. */
	if (chown(cache_name, pw->pw_uid, pw->pw_gid) == -1) {
	    LOG((LOG_CRIT, "pam_sm_setcred: chown: %s",
			strerror(errno)));
	    (void)krb5_cc_destroy(state->ctx, ccache_perm);
	}
	if (chmod(cache_name, (S_IRUSR|S_IWUSR)) == -1) {
	    LOG((LOG_CRIT, "pam_sm_setcred: chmod: %s",
			strerror(errno)));
	    (void)krb5_cc_destroy(state->ctx, ccache_perm);
	}
    }
    if (pamret == PAM_SUCCESS) {
    	(void)snprintf(buf, sizeof(buf), "KRB5CCNAME=%s:%s", cache_type,
	    cache_name);
	DBG(("pam_sm_setcred: %s", buf));
    	pamret = pam_putenv(state->pamh, buf);
	if (pamret != 0) {
	    DBG(("pam_sm_setcred: pam_putenv: %s",
		pam_strerror(state->pamh, pamret)));
	    (void)krb5_cc_destroy(state->ctx, ccache_perm);
	    pamret = PAM_SERVICE_ERR;
    	}
    	if (TEST_OPTION(state, OPT_PUTENV_DIRECT))
	    putenv(buf);
    }
    if (pamret == PAM_SUCCESS)
	state->ccache = ccache_perm;
    return pamret;
}

static void
expand_ccache_spec(const char *spec, struct passwd *pw, char **p)
{
    	char buf[PATH_MAX + 1];
    	char *q;
	const char *r, *s;
	size_t n, m;
	int overflow;

	overflow = 0;
	s = spec;
	q = buf;
	m = sizeof(buf) - 1;
	for (;;) {
	    r = strpbrk(s, "%\0");
	    n = r - s;
	    if (n > m) {
		overflow = 1;
		n = m;
	    }
	    if (n > 0)
	    	memcpy(q, s, n);
	    q += n;
	    m -= n;
	    if (m < 2)
		overflow = 1;
	    if (overflow || *r == '\0')
		break;
	    /* r[0] will be '%' here */
	    switch (r[1]) {
	    case 'u':
		n = snprintf(q, m, "%d", (int)pw->pw_uid);
		break;
	    case 'd':
		n = snprintf(q, m, "%d", (int)getpid());
		break;
	    default:
		n = 2;
		q[0] = '%';
		q[1] = r[1];
		break;
	    }
	    q += n;
	    m -= n;
	    if (n > m) {
		overflow = 1;
		break;
	    }
	    s = &r[2];
	}
	*q = '\0';
	if (overflow)
	    LOG((LOG_DEBUG,
		"pam_sm_setcred: overflow while processing ccache=%s", spec));
	*p = strdup(buf);
	if (*p == NULL)
	    LOG((LOG_DEBUG, "pam_sm_setcred: malloc (failure)"));
	return;
}

static int
delete_cred(struct pam_krb5_state *state)
{

    DBG(("pam_sm_cred: deleting credentials"));
    if (state->ccache) {
    	(void)krb5_cc_destroy(state->ctx, state->ccache);
	state->ccache = NULL;
    }
    return PAM_SUCCESS;
}

static int
reinitialize_cred(struct pam_krb5_state *state)
{

    DBG(("pam_sm_cred: reinitialize not implemented"));
    return PAM_SUCCESS;
}

static int
refresh_cred(struct pam_krb5_state *state)
{

    DBG(("pam_sm_cred: refresh not implemented"));
    return PAM_SUCCESS;
}
