/* vim:sw=4 ts=8 tw=0 noai cin cino=>1s+0.5s\:0t0
 *
 * pam_krb5_pass.c
 *
 * PAM password management functions for pam_krb5
 *
 */

static const char rcsid[] = "$Id: pam_krb5_pass.c,v 1.1.1.1.2.17 2001/09/24 20:19:45 nectar Exp $";

#include <errno.h>
#include <stdio.h>	/* sprintf */
#include <stdlib.h>	/* malloc */
#include <string.h>
#include <syslog.h>	/* syslog */

#include <krb5.h>
#include <com_err.h>
#include "pam_krb5.h"

static int	display_cpw_result(pam_handle_t *, int, krb5_data *,
		    krb5_data *);

/* Change a user's password */
int
pam_sm_chauthtok(pam_handle_t *pamh, int flags, int argc, const char **argv)
{
    krb5_error_code	 krbret;
    int			 result_code;
    krb5_data		 result_code_string, result_string;
    char	 	 prompt[80];
    int			 pamret;
    char		*pass = NULL, *pass2 = NULL;
    struct		 pam_krb5_state *state;
    krb5_creds		*creds;

    pamret = pam_krb5_initialize_method("pam_sm_chauthtok", pamh, &state,
	argc, argv);
    if (pamret != PAM_SUCCESS)
	return pamret;
    memset(&result_code_string, 0, sizeof(result_code_string));
    memset(&result_string, 0, sizeof(result_string));

    if (flags & PAM_UPDATE_AUTHTOK)
	DBG(("pam_sm_chauthtok: PAM_UPDATE_AUTHTOK"));
    else if (flags & PAM_PRELIM_CHECK) {
	DBG(("pam_sm_chauthtok: PAM_PRELIM_CHECK"));
	pamret = PAM_SUCCESS;
	goto final;
    } else {
	DBG(("pam_sm_chauthtok: bad flags"));
	pamret = PAM_AUTHTOK_ERR;
	goto final;
    } 

    if (state->chpw_creds == NULL) {
	DBG(("pam_sm_chauthtok: need to get chpw_creds"));
	pamret = pam_krb5_get_authtok(state, CRED_CHANGEPW,
	    &state->chpw_creds);
	if (pamret != PAM_SUCCESS)
	    goto final;
    } else
	DBG(("pam_sm_chauthtok: already have chpw_creds"));

    /* Now get the new password */
    if (pass == NULL) {
	(void)snprintf(prompt, sizeof(prompt), "Enter new password: ");
	if ((pamret = pam_krb5_get_user_info(pamh, prompt, PAM_PROMPT_ECHO_OFF,
	    &pass)) != 0) {
	    DBG(("pam_sm_chauthtok: pam_krb5_get_user_info(%s)",
		pam_strerror(pamh, pamret)));
	    pamret = PAM_SERVICE_ERR;
	    goto final;
	}
	(void)snprintf(prompt, sizeof(prompt), "Enter it again: ");
	if ((pamret = pam_krb5_get_user_info(pamh, prompt, PAM_PROMPT_ECHO_OFF,
	    &pass2)) != 0) {
	    DBG(("pam_sm_chauthtok: pam_krb5_get_user_info(%s)",
		pam_strerror(pamh, pamret)));
	    pamret = PAM_SERVICE_ERR;
	    goto final;
	}
	if (strcmp(pass, pass2) != 0) {
	    DBG(("pam_sm_chauthtok: passwords do not match"));
	    pamret = PAM_AUTHTOK_ERR;
	    goto final;
	}
    }

    /* Change it */
    krbret = krb5_change_password(state->ctx, state->chpw_creds, pass,
	&result_code, &result_code_string, &result_string);
    krb5_free_creds(state->ctx, state->chpw_creds);
    state->chpw_creds = NULL;
    if (krbret != 0) {	
	DBG(("pam_sm_chauthtok: krb5_change_password: %s",
	    error_message(krbret)));
	pamret = PAM_AUTHTOK_ERR;
	goto final;
    }
    pamret = display_cpw_result(state->pamh, result_code, &result_string,
	&result_code_string);
    if (result_code != 0)
	pamret = PAM_AUTHTOK_ERR;
    if (pamret != PAM_SUCCESS)
	goto final;
    pamret = pam_set_item(pamh, PAM_AUTHTOK, pass);
    if (pamret == PAM_SUCCESS && TEST_STATE(state, STATE_PW_EXPIRED)) {
	/*
	 * The password had been expired and no longer is.
	 * Now let's get our credentials.
	 */
	CLEAR_STATE(state, STATE_PW_EXPIRED);
	creds = NULL;
	pamret = pam_get_item(state->pamh, PAM_AUTHTOK,
	    (PAM_CONST void **)&pass);
	if (pamret == PAM_SUCCESS)
	    pamret = pam_krb5_get_cred(state, pass, CRED_TGT, &creds, NULL);
	if (pamret == PAM_SUCCESS) {
	    pamret = pam_krb5_store_tgt(state, creds);
	    krb5_free_creds(state->ctx, creds);
	    if (pamret == 0)
		pamret = pam_krb5_verify_tgt(state);
	    creds = NULL;
	}
	if (pamret == PAM_SUCCESS)
	    pamret = pam_krb5_acct(state);
    }

final:
    if (!(flags & PAM_PRELIM_CHECK) && state->chpw_creds != NULL) {
	krb5_free_creds(state->ctx, state->chpw_creds);
	state->chpw_creds = NULL;
    }
    FREE0(result_string.data);
    FREE0(result_code_string.data);
    pam_krb5_note_result("pam_sm_chauthtok", state, pamret);
    return pamret;
}

static int
display_cpw_result(pam_handle_t *pamh, int rc, krb5_data *result_string,
	krb5_data *result_code_string)
{
    const struct	pam_message *msgp;
    struct		pam_conv *conv;
    struct		pam_message msg;
    struct		pam_response *resp;
    char		buf[BUFSIZ];
    int			pamret, n, need_separator;

    pamret = pam_get_item(pamh, PAM_CONV, (PAM_CONST void **)&conv);
    if (pamret != 0) {
	LOG((LOG_CRIT, "display_cpw_result: no conversation: %s",
	    pam_strerror(pamh, pamret)));
	return PAM_SYSTEM_ERR;
    }
    need_separator = (result_code_string->length != 0) && 
	(result_string->length != 0);
    n = snprintf(buf, sizeof(buf), "%.*s%s%.*s\n",
	(int)result_code_string->length,
	(const char *)result_code_string->data,
	need_separator ? ": " : "",
	(int)result_string->length,
	(const char *)result_string->data);
    if (n >= sizeof(buf))
	buf[sizeof(buf)-2] = '\n';

    msg.msg       = buf;
    msg.msg_style = PAM_TEXT_INFO;
    msgp          = &msg;
    resp          = NULL;

    pamret = conv->conv(1, &msgp, &resp, (void *)conv->appdata_ptr);
    if (pamret != PAM_SUCCESS)
	pamret = PAM_AUTHTOK_ERR;
    FREE0(resp);
    return pamret;
}
