/* vim:sw=4 ts=8 tw=0 noai cin cino=>1s+0.5s\:0t0
 *
 * pam_krb5_prompter.c
 *
 * Kerberos 5 prompter
 *
 */
static const char rcsid[] = "$Id: pam_krb5_prompter.c,v 1.1.2.1 2001/09/24 14:58:15 nectar Exp $";

#include <sys/types.h>
#include <errno.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <com_err.h>
#include <krb5.h>
#include "pam_krb5.h"

krb5_error_code
pam_krb5_prompter(krb5_context context, void *data, const char *name,
	     const char *banner, int num_prompts, krb5_prompt prompts[])
{
    int		pam_prompts = num_prompts;
    int		pamret, i;

    const struct pam_message **conv_arg;
    struct pam_message	*msg, **p;
    struct pam_response	*resp = NULL;
    struct pam_conv	*conv;
    krb5_prompt_type	*ptypes = NULL;
    struct pam_krb5_state *state = (struct pam_krb5_state *)data;
    char *buf, *s;
    size_t buflen;

    pamret = pam_get_item(state->pamh, PAM_CONV, (PAM_CONST void **)&conv);
    if (pamret != 0) {
	DBG(("pam_krb5_prompter: couldn't get conversation function"));
	return KRB5KRB_ERR_GENERIC;
    }
    buflen = 0;
    if (TEST_OPTION(state, OPT_NO_TEXTINFO_PROMPTS)) {
	name = NULL;
	banner = NULL;
    }
    if (name != NULL) {
	buflen += strlen(name) + 1;
	pam_prompts++;
    }
    if (banner != NULL) {
	buflen += strlen(banner) + 1;
	pam_prompts++;
    }

    msg = calloc((size_t)pam_prompts, sizeof(*msg));
    if (msg == NULL)
	return ENOMEM;
    p = calloc((size_t)pam_prompts, sizeof(*p));
    if (p == NULL)
	return ENOMEM;
    for (i = 0; i < pam_prompts; i++)
	p[i] = &msg[i];

    /* Determine storage needed for the prompt strings.  This makes
     * all the following uses of sprintf safe, as well as simplifying
     * the storage management.
     */
    for (i = 0; i < num_prompts; i++)
	buflen += strlen(prompts[i].prompt) + 1;
    buf = malloc(buflen);
    if (buf == NULL)
	return ENOMEM;
    s = buf;

    /* Now use pam_prompts as an index */
    pam_prompts = 0;

    if (name) {
	msg[pam_prompts].msg = s;
	msg[pam_prompts].msg_style = PAM_TEXT_INFO;
	s += sprintf(s, "%s", name) + 1;
	DBG(("pam_krb5_prompter: PAM_TEXT_INFO `%s'", msg[pam_prompts].msg));
	pam_prompts++;
    }

    if (banner) {
	msg[pam_prompts].msg = s;
	msg[pam_prompts].msg_style = PAM_TEXT_INFO;
	s += sprintf(s, "%s", banner) + 1;
	DBG(("pam_krb5_prompter: PAM_TEXT_INFO `%s'", msg[pam_prompts].msg));
	pam_prompts++;
    }

    ptypes = krb5_get_prompt_types(context);

    for (i = 0; i < num_prompts; i++) {
	if (compat_is_new_password_prompt(ptypes, prompts, i) &&
	    TEST_STATE(state, STATE_SM_AUTHENTICATE) &&
	    !TEST_OPTION(state, OPT_PW_EXP_IN_AUTH)) {
	    free(buf);
	    free(msg);
	    free(p);
	    return KRB5KDC_ERR_KEY_EXP;
	}
	msg[pam_prompts].msg = strdup(prompts[i].prompt);
	msg[pam_prompts].msg_style = prompts[i].hidden ? PAM_PROMPT_ECHO_OFF
						       : PAM_PROMPT_ECHO_ON;
	s += sprintf(s, "%s", prompts[i].prompt) + 1;
	DBG(("pam_krb5_prompter: %s `%s'", 
	    prompts[i].hidden ? "PAM_PROMPT_ECHO_OFF" : "PAM_PROMPT_ECHO_ON",
	    msg[pam_prompts].msg));
	pam_prompts++;
    }

#if defined(PAM_LINUX_CODEBASE)
    conv_arg = (const struct pam_message **)p;
#else
    conv_arg = (const struct pam_message **)msg;
#endif
    pamret = conv->conv(pam_prompts, conv_arg, &resp, conv->appdata_ptr);
    if (pamret != 0) {
	DBG(("pam_krb5_prompter: conv returned %d", pamret));
	goto final;
    }

    if (!resp) {
	DBG(("pam_krb5_prompter: conv returned no response"));
	goto final;
    }

    /* Reuse pam_prompts as a starting index */
    pam_prompts = 0;
    if (name)
	pam_prompts++;
    if (banner)
	pam_prompts++;

    /* NOTE: When copying the PAM responses to the krb5 response data	*/
    /*       area we MUST copy a trailing null because			*/
    /*     krb5_get_init_creds_password() may use strcmp() to compare	*/
    /*     new passwords...						*/
    /*									*/
    /*     The length we stuff into the reply must not include the	*/
    /*     null however, otherwise the null may be included as part of	*/
    /*     a new password!!						*/
    for (i = 0; i < num_prompts; i++, pam_prompts++) {
	size_t len;
	if (!resp[pam_prompts].resp) {
	    pamret = PAM_AUTH_ERR;
	    DBG(("pam_krb5_prompter: resp %d is null", pam_prompts));
	    goto final;
	}
	len = strlen(resp[pam_prompts].resp); /* Help out the compiler */
	if (len > (size_t)prompts[i].reply->length) {
	    pamret = PAM_AUTH_ERR;
	    DBG(("pam_krb5_prompter: resp %d is too long", pam_prompts));
	    goto final;
	}
	memcpy(prompts[i].reply->data, resp[pam_prompts].resp, len + 1);
	prompts[i].reply->length = len;
    }

final:
    /* pam_prompts is correct at this point */

    free(buf);
    free(msg);
    free(p);

    if (resp) {
	for (i = 0; i < pam_prompts; i++) {
	    /*
	     * Note that PAM is underspecified wrt free()'ing resp[i].resp.
	     * It's not clear if I should free it, or if the application
	     * has to. Therefore most (all?) apps won't free() it, and I
	     * can't either, as I am not sure it was malloc()'d. All PAM
	     * implementations I've seen leak memory here. Not so bad, IFF
	     * you fork/exec for each PAM authentication (as is typical).
	     */
#if 0
	    if (resp[i].resp)
		free(resp[i].resp);
#endif /* 0 */
	}
	/* This does not lose resp[i].resp if the application saved a copy. */
	free(resp);
    }

    DBG(("pam_krb5_prompter: pamret=%d", pamret));

    return (pamret ? KRB5KRB_ERR_GENERIC : 0);
}
