/* vim:sw=4 ts=8 tw=0 noai cin cino=>1s+0.5s\:0t0
 * support.c
 *
 * Support functions for pam_krb5
 */

static const char rcsid[] = "$Id: support.c,v 1.1.1.1.2.14 2001/09/24 20:19:45 nectar Exp $";

#include <sys/types.h>	/* gid_t */
#include <errno.h>
#include <stdio.h>	/* BUFSIZ */
#include <stdlib.h>	/* malloc */
#include <string.h>	/* strncpy */
#include <syslog.h>	/* syslog */
#include <pwd.h>
#include <stdarg.h>
#include <unistd.h>	/* getuid */
#include <krb5.h>
#include <com_err.h>
#include "pam_krb5.h"

typedef struct pam_cred {
    uid_t cr_uid;
    uid_t cr_ruid;
    gid_t cr_gid;
    gid_t cr_rgid;
} cred_t;

static int	set_options(struct pam_krb5_options *, int, const char **);
static void	cleanup_state(pam_handle_t *, void *, int);
static void	dump_state(struct pam_krb5_state *);
static int	get_user_and_service(struct pam_krb5_state *);

int
pam_krb5_initialize_method(const char *method, pam_handle_t *pamh,
    struct pam_krb5_state **state, int argc, const char *argv[])
{
    int		pamret;
    struct	pam_krb5_state *s;
    krb5_error_code krbret;

    LOG((LOG_DEBUG, "initialize_method: %s", method));
    s = NULL;
    pamret = pam_get_data(pamh, "pam_krb5_state", (PAM_CONST void **)state);
    if (pamret != PAM_SUCCESS) {
	initialize_krb5_error_table();
	initialize_asn1_error_table();
	LOG((LOG_DEBUG,
	    "initialize_method: allocating pam_krb5_state"));
	s = calloc(1, sizeof(*s));
	if (s == NULL) {
	    LOG((LOG_CRIT, "initialize_method: out of memory"));
	    pamret = PAM_SYSTEM_ERR;
	    goto fail;
	}
	pamret = pam_set_data(pamh, "pam_krb5_state", s, cleanup_state);
	if (pamret != PAM_SUCCESS) {
	    LOG((LOG_CRIT, "initialize_method: pam_set_data: %s",
		pam_strerror(pamh, pamret)));
	    goto fail;
	}
	krbret = krb5_init_context(&s->ctx);
	if (krbret != 0) {
	    LOG((LOG_CRIT, "initialize_method: krb5_init_context: %s",
		error_message(krbret)));
	    pamret = PAM_SERVICE_ERR;
	    goto fail;
	}
	pamret = set_options(&s->opt, argc, argv);
	if (pamret != PAM_SUCCESS)
	    goto fail;
	*state = s;
    }
    (*state)->pamh = pamh;
    pamret = get_user_and_service(*state);
    if (pamret != PAM_SUCCESS)
	goto fail;
    if (TEST_OPTION(*state, OPT_DEBUG))
	dump_state(*state);
    (*state)->method = method;
    LOG((LOG_DEBUG, "initialize_method: success"));
    return PAM_SUCCESS;

fail:
    LOG((LOG_DEBUG, "initialize_method: failure"));
    FREE0(s);
    return pamret;
}

static void
dump_state(struct pam_krb5_state *state)
{
    static struct { const char *name; unsigned long val; } bool_opts[] = {
#define OPTION(name, flag, val) { name, flag },
#include "option_flags.h"
#undef OPTION
    };
    int i;
    DBG(("dumping state"));
    for (i = 0; i < sizeof(bool_opts)/sizeof(bool_opts[0]); i++) {
	if (TEST_OPTION(state, bool_opts[i].val))
	    DBG(("option: %s", bool_opts[i].name));
    }
#define OPTION(name, member) \
    if (state->opt.member != NULL) \
	DBG(("option: %s%s", name, state->opt.member));
#include "option_strings.h"
#undef OPTION
    if (TEST_STATE(state, STATE_WILL_IGNORE))
	DBG(("state: STATE_WILL_IGNORE"));
    if (TEST_STATE(state, STATE_SM_AUTHENTICATE))
	DBG(("state: STATE_SM_AUTHENTICATE"));
    if (TEST_STATE(state, STATE_PW_EXPIRED))
	DBG(("state: STATE_PW_EXPIRED"));
    if (TEST_STATE(state, STATE_AUTH_COMPLETED))
	DBG(("state: STATE_AUTH_COMPLETED"));
    if (state->princ_name != NULL)
	DBG(("state: princ_name=`%s'", state->princ_name));
    if (state->user != NULL)
	DBG(("state: user=`%s'", state->user));
    if (state->service != NULL)
	DBG(("state: service=`%s'", state->service));
    if (state->c_ucred != NULL)
	DBG(("state: c_ucred exists"));
    if (state->princ != NULL)
	DBG(("state: princ exists"));
    if (state->ccache != NULL)
	DBG(("state: ccache exists"));
    if (state->chpw_creds != NULL)
	DBG(("state: chpw_creds exists"));
}

void
pam_krb5_note_result(const char *method, struct pam_krb5_state *state,
    int pamret)
{
    const char *p;
    if (pamret == PAM_USER_UNKNOWN || state->user == NULL)
	p = "(unknown user)";
    else
	p = state->user;
    LOG((LOG_NOTICE, "%s: result for user `%s': %s", method, p,
	pam_strerror(state->pamh, pamret)));
}

static int
get_user_and_service(struct pam_krb5_state *state)
{
    int	e;

    e = pam_get_user(state->pamh, (PAM_CONST char **)&state->user, "login: ");
    if (e != PAM_SUCCESS)
	LOG((LOG_CRIT, "PAM_USER failed: %s", pam_strerror(state->pamh, e)));
    else {
	e = pam_get_item(state->pamh, PAM_SERVICE,
	    (PAM_CONST void **)&state->service);
	if (e != PAM_SUCCESS)
	    state->service = "unknown";
	e = PAM_SUCCESS;
    }
    return e;
}

static void
cleanup_state(pam_handle_t *pamh, void *data, int pam_end_status)
{

    LOG((LOG_DEBUG, "cleanup_state"));
    if (data != NULL) {
	struct pam_krb5_state *state;

	state = (struct pam_krb5_state *)data;
	FREE0(state->princ_name);
	FREE0(state->c_ucred);
	if (state->princ)
		krb5_free_principal(state->ctx, state->princ);
	if (state->ccache)
		krb5_cc_close(state->ctx, state->ccache);
	if (state->chpw_creds)
		krb5_free_creds(state->ctx, state->chpw_creds);
	krb5_free_context(state->ctx);
	FREE0(state->opt.ccache_spec);
	FREE0(state->opt.renewable);
	FREE0(state);
    }
}

int
pam_krb5_become_user(struct pam_krb5_state *state, struct passwd *pw)
{
    int		 ret = 0, pret = PAM_SUCCESS;

    if (pw == NULL)
	return PAM_SERVICE_ERR;
    state->c_ucred = malloc(sizeof(*state->c_ucred));
    if (state->c_ucred == NULL)
	return PAM_SYSTEM_ERR;

    /* Caller's Unix creds */
    state->c_ucred->cr_ruid=getuid(); /* real */
    state->c_ucred->cr_rgid=getgid(); /* real */
    state->c_ucred->cr_uid=geteuid(); /* effective */
    state->c_ucred->cr_gid=getegid(); /* effective */

    DBG(("pam_krb5_become_user: start uid=%lu, euid=%lu, gid=%lu, egid=%lu",
	(unsigned long)state->c_ucred->cr_ruid,
	(unsigned long)state->c_ucred->cr_uid,
	(unsigned long)state->c_ucred->cr_rgid,
	(unsigned long)state->c_ucred->cr_gid));
    /* Check if we're root or the user already */
    if (state->c_ucred->cr_ruid != 0 && state->c_ucred->cr_uid != 0) {
	/* XXX The following is broken for applications that run as a
	 * special user ID (e.g. `www' for web servers).
	 */
	/* If we're neither root nor the user already, then give up */
	if (state->c_ucred->cr_ruid != pw->pw_uid && 
	    state->c_ucred->cr_uid != pw->pw_uid) {
	    DBG(("pam_krb5_become_user: not root nor user `%s'", pw->pw_name));
	    FREE0(state->c_ucred);
	    return PAM_PERM_DENIED;
	}

	/* We're the user already */
	if (state->c_ucred->cr_uid == pw->pw_uid) {
	    DBG(("pam_krb5_become_user: already user `%s'", pw->pw_name));
	    FREE0(state->c_ucred);
	    return PAM_SUCCESS;
	}
    }

    if (state->c_ucred->cr_gid != pw->pw_gid) {
	if ((ret = setregid(state->c_ucred->cr_gid, pw->pw_gid)) != 0) {
	    DBG(("pam_krb5_become_user: setregid: %s", strerror(errno)));
	    FREE0(state->c_ucred);
	    return PAM_PERM_DENIED;
	}
    }
    if (state->c_ucred->cr_uid != pw->pw_uid) {
	if ((ret = setreuid(state->c_ucred->cr_uid, pw->pw_uid)) != 0) {
	    DBG(("pam_krb5_become_user: setreuid: %s", strerror(errno)));
	    if (pret == PAM_PERM_DENIED)
		(void)setregid(state->c_ucred->cr_rgid,
		    state->c_ucred->cr_gid);
	    FREE0(state->c_ucred);
	    return PAM_PERM_DENIED;
	}
    }
    DBG(("pam_krb5_become_user: end uid=%lu, euid=%lu, gid=%lu, egid=%lu",
	(unsigned long)getuid(), (unsigned long)geteuid(),
	(unsigned long)getgid(), (unsigned long)getegid()));
    return PAM_SUCCESS;
}

int
pam_krb5_unbecome_user(struct pam_krb5_state *state)
{
    int		ret = 0, pret = PAM_SUCCESS;

    if (state->c_ucred == NULL)
	return PAM_SUCCESS;

    /* Reset gid/egid to pam_krb5_become_user()'s caller's values */
    DBG(("pam_krb5_unbecome_user: start uid=%lu, euid=%lu, gid=%lu, egid=%lu",
	(unsigned long)getuid(), (unsigned long)geteuid(),
	(unsigned long)getgid(), (unsigned long)getegid()));
    if ((ret = setregid(state->c_ucred->cr_rgid,
	state->c_ucred->cr_gid)) != 0) {
	DBG(("pam_krb5_unbecome_user: setregid: %s", strerror(errno)));
	pret = PAM_SYSTEM_ERR;
    }

    /* Reset uid/euid to become_user()'s caller's values */
    if ((ret = setreuid(state->c_ucred->cr_ruid,
	state->c_ucred->cr_uid)) != 0) {
	DBG(("pam_krb5_unbecome_user: setreuid: %s", strerror(errno)));
	pret = PAM_SYSTEM_ERR;
    }
    DBG(("pam_krb5_unbecome_user: end uid=%lu, euid=%lu, gid=%lu, egid=%lu",
	(unsigned long)getuid(), (unsigned long)geteuid(),
	(unsigned long)getgid(), (unsigned long)getegid()));
    DBG(("pam_krb5_unbecome_user: returning %d", pret));
    FREE0(state->c_ucred);
    return pret;
}

/*
 * Get info from the user. Disallow null responses (regardless of flags).
 * response gets allocated and filled in on successful return. Caller
 * is responsible for freeing it.
 */
int
pam_krb5_get_user_info(pam_handle_t *pamh, char *prompt, int type,
    char **response)
{
    int pamret;
    struct pam_message	msg;
    const struct pam_message *pmsg;
    struct pam_response	*resp = NULL;
    struct pam_conv	*conv;

    if ((pamret = pam_get_item(pamh, PAM_CONV, (PAM_CONST void **) &conv)) != 0)
	return pamret;

    /* set up conversation call */
    pmsg = &msg;
    msg.msg_style = type;
    msg.msg = prompt;

    if ((pamret = conv->conv(1, (PAM_CONST struct pam_message **) &pmsg,
				&resp, conv->appdata_ptr)) != 0)
	return pamret;

    /* Caller should ignore errors for non-response conversations */
    if (!resp)
	return PAM_CONV_ERR;

    if (!(resp->resp && resp->resp[0])) {
	free(resp);
	return PAM_AUTH_ERR;
    }

    *response = resp->resp;
    free(resp);
    return pamret;
}

/*
 * This routine with some modification is from the MIT V5B6 appl/bsd/login.c
 * Modified by Sam Hartman <hartmans@mit.edu> to support PAM services
 * for Debian.
 *
 * Verify the Kerberos ticket-granting ticket just retrieved for the
 * user.  If the Kerberos server doesn't respond, assume the user is
 * trying to fake us out (since we DID just get a TGT from what is
 * supposedly our KDC) and return PAM_AUTH_ERR.  If we have no keytab
 * entry with which to perform the validation, then depending upon option
 * settings, return one of the following: if `missing_keytab_ok' is set,
 * return PAM_SUCCESS; if `require_keytab' is set, return
 * PAM_AUTH_ERR; otherwise (if neither option is set), return PAM_IGNORE.
 *
 * And of course, if we have a keytab, return PAM_SUCCESS or PAM_AUTH_ERR
 * depending upon the result of the verification.
 *
 */
int
pam_krb5_verify_tgt(struct pam_krb5_state *state)
{
    char		 phost[BUFSIZ];
    const char		*services [3];
    const char		**service;
    krb5_error_code	 retval = -1;
    krb5_principal	 princ;
    krb5_keyblock	*keyblock = 0;
    krb5_data		 packet;
    krb5_auth_context	 auth_context = NULL;
    int			 epam;

    packet.data = 0;

    /*
     * If possible we want to try and verify the ticket we have
     * received against a keytab.  We will try multiple service
     * principals, including at least the host principal and the PAM
     * service principal.  The host principal is preferred because access
     * to that key is generally sufficient to compromise root, while the
     *     service key for this PAM service may be less carefully guarded.
     * It is important to check the keytab first before the KDC so we do
     * not get spoofed by a fake  KDC.
     */
    services [0] = "host";
    services [1] = state->service;
    services [2] = NULL;
    for (service = &services[0]; *service != NULL; service++) {
      if ((retval = krb5_sname_to_principal(state->ctx, NULL, *service,
	KRB5_NT_SRV_HST, &princ)) != 0) {
	DBG(("pam_krb5_verify_tgt: krb5_sname_to_principal: %s",
	    error_message(retval)));
	return PAM_SERVICE_ERR;
      }

      /* Extract the name directly. */
      strncpy(phost, compat_princ_component(state->ctx, princ, 1),
        sizeof(phost));
      phost[sizeof(phost) - 1] = '\0';

      /*
       * Do we have service/<host> keys?
       * (use default/configured keytab, kvno IGNORE_VNO to get the
       * first match, and ignore enctype.)
       */
      if ((retval = krb5_kt_read_service_key(state->ctx, NULL, princ, 0,
        0, &keyblock)) != 0)
	continue;
      else
        break;
    }
    if (retval != 0) {		/* failed to find key */
	/* Keytab or service key does not exist */
	DBG(("pam_krb5_verify_tgt: krb5_kt_read_service_key: %s",
	    error_message(retval)));
	if (TEST_OPTION(state, OPT_REQUIRE_KEYTAB))
	    epam = PAM_AUTH_ERR;
	else if (TEST_OPTION(state, OPT_MISSING_KEYTAB_OK))
	    epam = PAM_SUCCESS;
	else
	    epam = PAM_IGNORE;
	goto cleanup;
    }
    if (keyblock)
	krb5_free_keyblock(state->ctx, keyblock);

    /* Talk to the kdc and construct the ticket. */
    retval = krb5_mk_req(state->ctx, &auth_context, 0, *service, phost,
			 NULL, state->ccache, &packet);
    if (auth_context) {
	krb5_auth_con_free(state->ctx, auth_context);
	auth_context = NULL; /* setup for rd_req */
    }
    if (retval != 0) {
	DBG(("pam_krb5_verify_tgt: krb5_mk_req: %s", error_message(retval)));
	(void)krb5_cc_destroy(state->ctx, state->ccache);
	state->ccache = NULL;
	epam = PAM_AUTH_ERR;
	goto cleanup;
    }

    /* Try to use the ticket. */
    retval = krb5_rd_req(state->ctx, &auth_context, &packet, princ,
			 NULL, NULL, NULL);
    if (retval) {
	DBG(("pam_krb5_verify_tgt: krb5_rd_req: %s", error_message(retval)));
	epam = PAM_AUTH_ERR;
    } else 
	epam = PAM_SUCCESS;

cleanup:
    if (packet.data)
	krb5_free_data_contents(state->ctx, &packet);
    krb5_free_principal(state->ctx, princ);

    DBG(("pam_krb5_verify_tgt: %s", pam_strerror(state->pamh, epam)));
    return epam;
}


/* Free the memory for cache_name. Called by pam_end() */
void
pam_krb5_cleanup_ccache(pam_handle_t *pamh, void *data, int pam_end_status)
{
    krb5_context	pam_context;
    krb5_ccache		ccache;
    krb5_error_code	retval;

    LOG((LOG_DEBUG, "cleanup_ccache: entry"));
    if ((retval = krb5_init_context(&pam_context)) != 0) {
	LOG((LOG_DEBUG, "cleanup_ccache: krb5_init_context: %s",
	    error_message(retval)));
	return;
    }
    ccache = (krb5_ccache) data;
    (void) krb5_cc_destroy(pam_context, ccache);
    krb5_free_context(pam_context);
    LOG((LOG_DEBUG, "cleanup_ccache: exit"));
}

void
pam_krb5_cleanup_creds(pam_handle_t *pamh, void *data, int pam_end_status)
{
    krb5_context	pam_context;
    krb5_error_code	retval;

    LOG((LOG_DEBUG, "pam_krb5_cleanup_creds: entry"));
    if ((retval = krb5_init_context(&pam_context)) != 0) {
	LOG((LOG_DEBUG, "pam_krb5_cleanup_creds: krb5_init_context: %s",
	    error_message(retval)));
	return;
    }

    ((krb5_creds *) data)->client = NULL;
    ((krb5_creds *) data)->server = NULL;
    (void) krb5_free_cred_contents(pam_context, (krb5_creds *) data);
    krb5_free_context(pam_context);
    LOG((LOG_DEBUG, "pam_krb5_cleanup_creds: exit"));
}


/* syslogging function for debug messages */
void 
pam_krb5_debug(const char *format, ...)
{
    char buf[128];
    static char prefix[] = "(pam_krb5) ";
    va_list args;

    snprintf(buf, sizeof(buf), "%s%s", prefix, format);
    va_start(args, format);
    vsyslog(LOG_DEBUG|LOG_AUTH, buf, args);
    va_end(args);
}

/* syslogging function for log messages */
void 
pam_krb5_log(int priority, const char *format, ...)
{
    char buf[128];
    static char prefix[] = "(pam_krb5) ";
    va_list args;

    snprintf(buf, sizeof(buf), "%s%s", prefix, format);
    va_start(args, format);
    vsyslog(priority|LOG_AUTH, buf, args);
    va_end(args);
}


static int
set_options(struct pam_krb5_options *opt, int ac, const char *av[])
{
    static struct { const char *name; unsigned long val; } bool_opts[] = {
#define OPTION(name, flag, val) \
	{ name, flag },
#include "option_flags.h"
#undef OPTION
    };
    int		 i, j, ok;
    char	*p;

    memset(opt, 0, sizeof(*opt));
    if (opt == NULL) {
	LOG((LOG_CRIT, "pam_krb5_set_options: out of memory"));
	return PAM_BUF_ERR;
    }
    for (i = 0; i < ac; i++) {
	ok = 0;
    	for (j = 0; j < sizeof(bool_opts) / sizeof(bool_opts[0]); j++) {
	    if (strcmp(av[i], bool_opts[j].name) == 0) {
		opt->flags |= bool_opts[j].val;
		ok = 1;
		break;
	    }
#define OPTION(name, member) \
	    p = strchr(av[i], '='); \
	    if (p != NULL && strncmp(av[i], name, p - av[i]) == 0) { \
		p = strdup(p + 1); \
		if (p == NULL) { \
		    LOG((LOG_CRIT, \
			"pam_krb5_set_options: out of memory")); \
		    return PAM_BUF_ERR; \
		} \
		opt->member = p; \
		ok = 1; \
		break; \
	    }
#include "option_strings.h"
#undef OPTION
    	}
	if (!ok)
	    LOG((LOG_DEBUG, "unknown option `%s'", av[i]));
    }
    return PAM_SUCCESS;
}
